"""
Ilab parsers
============

Parsers provided by this module include:

IlabModuleList - command ``/usr/bin/ilab model list``
-----------------------------------------------------
"""
from insights.core import CommandParser
from insights.core.plugins import parser
from insights.core.exceptions import SkipComponent, ParseException
from insights.specs import Specs


@parser(Specs.ilab_model_list)
class IlabModuleList(CommandParser, list):
    """
    This parser will parse the output of "/usr/bin/ilab model list".

    Sample output from ``/usr/bin/ilab model list``::

        +-----------------------------------+---------------------+---------+
        | Model Name                        | Last Modified       | Size    |
        +-----------------------------------+---------------------+---------+
        | models/prometheus-8x7b-v2-0       | 2024-08-09 13:28:50 |  87.0 GB|
        | models/mixtral-8x7b-instruct-v0-1 | 2024-08-09 13:28:24 |  87.0 GB|
        | models/granite-7b-redhat-lab      | 2024-08-09 14:28:40 |  12.6 GB|
        | models/granite-7b-starter         | 2024-08-09 14:40:35 |  12.6 GB|
        +-----------------------------------+---------------------+---------+

    Examples:
        >>> type(ilab_model_list_info)
        <class 'insights.parsers.ilab.IlabModuleList'>
        >>> ilab_model_list_info[0]["model_name"]
        'models/prometheus-8x7b-v2-0'

    Attributes:
        models (list): list of model names

    """
    def parse_content(self, content):
        if not content:
            raise SkipComponent("Empty")
        if not content[0].startswith("+-"):
            raise ParseException("Unexpected format: %s" % content[0])
        for line in content:
            if line.startswith("|") and "Model Name" not in line:
                split_items = line.split("|")
                if len(split_items) == 5:
                    self.append(
                        {
                            "model_name": split_items[1].strip(),
                            "last_modified": split_items[2].strip(),
                            "size": split_items[3].strip(),
                        }
                    )
                else:
                    raise ParseException("Unexpected format: %s" % line)
        self.models = [m['model_name'] for m in self]



"""
Filtering
=========
"""

import logging

logger = logging.getLogger(__name__)


class AllowFilter(object):
    """
    Class for filtering per allow list.
    """

    def parse_line(self, line, **kwargs):
        # filter line as per the allow list specified by plugins
        if not line:
            return line
        allowlist = kwargs.get('allowlist', {})
        if allowlist:
            for a_key in list(allowlist.keys()):
                # keep line when any filter match
                # FIXME:
                # Considering performance, didn't handle multiple filters in one same line
                if a_key in line:
                    allowlist[a_key] -= 1
                    # stop checking it when enough lines contain the key were found
                    allowlist.pop(a_key) if allowlist[a_key] == 0 else None
                    return line
        # discard line when none filters found

    def generate_report(self, report_dir, archive_name):
        pass  # pragma: no cover



"""
IP Obfuscation
==============
The following cleaners are included in this module:

IPv4 Obfuscation
----------------

IPv6 Obfuscation
----------------

"""

import logging
import os
import re
import socket
import struct

from insights.cleaner.utilities import write_report

logger = logging.getLogger(__name__)


class IPv4(object):
    """
    Class for obfuscating IPv4.
    """

    def __init__(self):
        # - IP obfuscate information
        self._ip_db = dict()  # IP database
        self._start_ip = '10.230.230.1'
        self._ignore_list = ["127.0.0.1"]

    def _ip2int(self, ipstr):
        # converts a dotted decimal IP address into an integer that can be incremented
        return struct.unpack('!I', socket.inet_aton(ipstr))[0]

    def _int2ip(self, num):
        # converts an integer stored in the IP database into a dotted decimal IP
        return socket.inet_ntoa(struct.pack('!I', num))

    def _ip2db(self, ip):
        '''
        adds an IP address to the IP database and returns the obfuscated entry, or returns the
        existing obfuscated IP entry
        FORMAT:
        {$obfuscated_ip: $original_ip,}
        '''
        ip_num = self._ip2int(ip)
        ip_found = False
        db = self._ip_db
        for k, v in db.items():
            if v == ip_num:
                ret_ip = self._int2ip(k)
                ip_found = True
        if ip_found:  # the entry already existed
            return ret_ip
        else:  # the entry did not already exist
            if len(self._ip_db) > 0:
                new_ip = max(db.keys()) + 1
            else:
                new_ip = self._ip2int(self._start_ip)
            db[new_ip] = ip_num

            return self._int2ip(new_ip)

    def parse_line(self, line, **kwargs):
        '''
        This will substitute an obfuscated IP for each instance of a given IP in a file
        It scans a given line and if an IP exists, it obfuscates the IP using _ip2db and returns the altered line
        '''

        def _sub_ip(line, ip):
            new_ip = self._ip2db(ip)
            logger.debug("Obfuscating IP - %s > %s", ip, new_ip)
            return line.replace(ip, new_ip)

        def _sub_ip_keep_width(line, ip):
            ip_len = len(ip)
            new_ip = self._ip2db(ip)
            new_ip_len = len(new_ip)
            logger.debug("Obfuscating IP - %s > %s", ip, new_ip)
            # pad or remove spaces to allow for the new length
            if ip_len > new_ip_len:
                numspaces = ip_len - new_ip_len
                line = line.replace(ip, new_ip)
                # shift past port specification to add spaces
                idx = line.index(new_ip) + new_ip_len
                c = line[idx]
                while c != " ":
                    idx += 1
                    if idx == len(line):
                        idx = len(line) - 1
                        break
                    c = line[idx]
                return line[0:idx] + numspaces * " " + line[idx:]
            elif new_ip_len > ip_len:
                numspaces = new_ip_len - ip_len
                line = line.replace(ip, new_ip)
                # shift past port specification to skip spaces
                idx = line.index(new_ip) + new_ip_len
                c = line[idx]
                while c != " ":
                    idx += 1
                    if idx == len(line):
                        break
                    c = line[idx]
                return line[0:idx] + line[(idx + numspaces) :]
            else:
                return line.replace(ip, new_ip)

        if not line:
            return line
        try:
            pattern = r"(((\b25[0-5]|\b2[0-4][0-9]|\b1[0-9][0-9]|\b[1-9][0-9]|\b[1-9]))(\.(\b25[0-5]|\b2[0-4][0-9]|\b1[0-9][0-9]|\b[1-9][0-9]|\b[0-9])){3})"
            ips = [each[0] for each in re.findall(pattern, line)]
            if len(ips) > 0:
                for ip in sorted(ips, key=len, reverse=True):
                    if ip not in self._ignore_list and ip in line:
                        if kwargs.get('width', False):
                            line = _sub_ip_keep_width(line, ip)
                        else:
                            line = _sub_ip(line, ip)
            return line
        except Exception as e:  # pragma: no cover
            logger.warning(e)
            raise Exception('SubIPError: Unable to Substitute IP Address - %s', ips)

    def mapping(self):
        mapping = []
        for k, v in self._ip_db.items():
            mapping.append({'original': self._int2ip(v), 'obfuscated': self._int2ip(k)})
        return mapping

    def generate_report(self, report_dir, archive_name):
        try:
            ip_report_file = os.path.join(report_dir, "%s-ip.csv" % archive_name)
            logger.info('Creating IP Report - %s', ip_report_file)
            lines = ['Obfuscated IP,Original IP']
            for k, v in self._ip_db.items():
                lines.append('{0},{1}'.format(self._int2ip(k), self._int2ip(v)))
        except Exception as e:  # pragma: no cover
            logger.exception(e)
            raise Exception('CreateReport Error: Error Creating IP Report')

        write_report(lines, ip_report_file)

        logger.info('Completed IP Report.')


# TODO
class IPv6(object):
    """
    Class for obfuscating IPv6.
    """

    pass



#!/usr/bin/env python
"""
This module runs insights and serializes the results into a directory. It is
configurable with a yaml manifest that specifies what to load, what to run,
and what to serialize. If a manifest isn't provided, a default one is used that
runs all datasources in ``insights.specs.Specs`` and
``insights.specs.default.DefaultSpecs`` and saves all datasources in
``insights.specs.Specs``.
"""
from __future__ import print_function

import argparse
import logging
import os
import sys
import tempfile
from datetime import datetime

import yaml

from insights import apply_configs
from insights import apply_default_enabled
from insights import get_pool
from insights.cleaner import Cleaner
from insights.core import blacklist
from insights.core import dr
from insights.core import filters
from insights.core.serde import Hydration
from insights.util import fs
from insights.util import utc
from insights.util.hostname import determine_hostname
from insights.util.subproc import call

SAFE_ENV = {
    "PATH": os.path.pathsep.join(
        [
            "/bin",
            "/usr/bin",
            "/sbin",
            "/usr/sbin",
            "/usr/share/Modules/bin",
        ]
    ),
    "LC_ALL": "C",
}

if "LANG" in os.environ:
    SAFE_ENV["LANG"] = os.environ["LANG"]

log = logging.getLogger(__name__)

default_manifest = """
---
# version is for the format of this file, not its contents.
version: 0

client:
  context:
    class: insights.core.context.HostContext
    args:
      timeout: 10 # timeout in seconds for commands. Doesn't apply to files.

  # commands and files to ignore
  blacklist:
    files: []
    commands: []
    patterns: []
    keywords: []

  # Can be a list of dictionaries with name/enabled fields or a list of strings
  # where the string is the name and enabled is assumed to be true. Matching is
  # by prefix, and later entries override previous ones. Persistence for a
  # component is disabled by default.
  persist:
    - name: insights.specs.Specs
      enabled: true

  run_strategy:
    name: serial
    args:
      max_workers: null

plugins:
  # disable everything by default
  # defaults to false if not specified.
  default_component_enabled: false

  # packages and modules to load
  packages:
    - insights.specs.default
    - insights.specs.datasources

  # configuration of loaded components. names are prefixes, so any component with
  # a fully qualified name that starts with a key will get the associated
  # configuration applied. Can specify timeout, which will apply to command
  # datasources. Can specify metadata, which must be a dictionary and will be
  # merged with the components' default metadata.
  configs:
    - name: insights.specs.Specs
      enabled: true
    - name: insights.specs.default.DefaultSpecs
      enabled: true
    - name: insights.specs.datasources
      enabled: true
    # needed for specs that aren't given names before used in DefaultSpecs
    - name: insights.core.spec_factory
      enabled: true

    # needed by multiple datasource specs/components
    - name: insights.parsers.hostname.Hostname
      enabled: true
    - name: insights.parsers.hostname.HostnameDefault
      enabled: true
    - name: insights.combiners.hostname.Hostname
      enabled: true
    - name: insights.parsers.uname.Uname
      enabled: true
    - name: insights.parsers.redhat_release.RedhatRelease
      enabled: true
    - name: insights.combiners.redhat_release.RedHatRelease
      enabled: true
    - name: insights.parsers.ps.PsAuxcww
      enabled: true
    - name: insights.parsers.ps.PsAuxww
      enabled: true
    - name: insights.combiners.ps.Ps
      enabled: true
    - name: insights.parsers.dmidecode.DMIDecode
      enabled: true
    - name: insights.parsers.installed_rpms.InstalledRpms
      enabled: true
    - name: insights.parsers.mount.ProcMounts
      enabled: true

    # needed for identifying RHEL major version
    - name: insights.components.rhel_version.IsRhel6
      enabled: true
    - name: insights.components.rhel_version.IsRhel7
      enabled: true
    - name: insights.components.rhel_version.IsRhel8
      enabled: true
    - name: insights.components.rhel_version.IsRhel9
      enabled: true

    # needed for cloud specs
    - name: insights.parsers.yum.YumRepoList
      enabled: true
    - name: insights.parsers.rhsm_conf.RHSMConf
      enabled: true
    - name: insights.combiners.cloud_provider.CloudProvider
      enabled: true
    - name: insights.components.cloud_provider.IsAWS
      enabled: true
    - name: insights.components.cloud_provider.IsAzure
      enabled: true
    - name: insights.components.cloud_provider.IsGCP
      enabled: true

    # needed for ceph specs
    - name: insights.components.ceph.IsCephMonitor
      enabled: true

    # needed for pcp specs
    - name: insights.parsers.systemd.unitfiles.UnitFiles
      enabled: true
    - name: insights.parsers.ros_config.RosConfig
      enabled: true

    # needed for 'teamdctl_config/state_dump' spec and nmcli_conn_show_uuids spec
    - name: insights.parsers.nmcli.NmcliConnShow
      enabled: true

    # needed for mssql_tls_cert_enddate
    - name: insights.parsers.mssql_conf.MsSQLConf
      enabled: true

    # needed for rsyslog_tls_cert_file
    - name: insights.parsers.rsyslog_conf.RsyslogConf
      enabled: true
    - name: insights.combiners.rsyslog_confs.RsyslogAllConf
      enabled: true

    # needed for sap specs
    - name: insights.parsers.saphostctrl.SAPHostCtrlInstances
      enabled: true
    - name: insights.combiners.sap.Sap
      enabled: true

    # needed for fw_security specs
    - name: insights.parsers.virt_what.VirtWhat
      enabled: true
    - name: insights.combiners.virt_what.VirtWhat
      enabled: true
    - name: insights.components.virtualization.IsBareMetal
      enabled: true

    # needed for 'modinfo_filtered_modules' spec
    - name: insights.parsers.lsmod.LsMod
      enabled: true

    # needed for satellite server specs
    - name: insights.combiners.satellite_version.SatelliteVersion
      enabled: true
    - name: insights.combiners.satellite_version.CapsuleVersion
      enabled: true
    - name: insights.components.satellite.IsSatellite
      enabled: true
    - name: insights.components.satellite.IsSatellite614AndLater
      enabled: true
    - name: insights.components.satellite.IsSatelliteLessThan614
      enabled: true
    - name: insights.components.satellite.IsSatellite611
      enabled: true
    - name: insights.components.satellite.IsCapsule
      enabled: true

    # needed for container specs
    - name: insights.parsers.podman_list.PodmanListContainers
      enabled: true
    - name: insights.parsers.docker_list.DockerListContainers
      enabled: true

    # needed for 'luks_data_sources' spec
    - name: insights.parsers.blkid.BlockIDInfo
      enabled: true
    - name: insights.components.cryptsetup.HasCryptsetupWithTokens
      enabled: true
    - name: insights.components.cryptsetup.HasCryptsetupWithoutTokens
      enabled: true

    # needed for 'iris' specs
    - name: insights.parsers.iris.IrisList
      enabled: true
    - name: insights.parsers.iris.IrisCpf
      enabled: true

    # needed for ausearch_insights_client
    - name: insights.components.rhel_version.IsGtOrRhel86
      enabled: true

    # needed for sealert spec
    - name: insights.parsers.selinux_config.SelinuxConfig
      enabled: true
    - name: insights.components.selinux.SELinuxEnabled
      enabled: true

    # needed for the 'fstab_mounted.dirs' to the 'ls_lan' spec
    - name: insights.parsers.fstab.FSTab
      enabled: true
""".strip()

EXCEPTIONS_TO_REPORT = set([OSError])
"""Exception types that should be reported on after core collection."""


def load_manifest(data):
    """Helper for loading a manifest yaml doc."""
    if isinstance(data, dict):
        return data
    if os.path.isfile(data):
        with open(data, 'r') as f:
            doc = yaml.safe_load(f)
    else:
        doc = yaml.safe_load(data)
    if not isinstance(doc, dict):
        raise Exception("Manifest didn't result in dict.")
    return doc


def load_packages(pkgs):
    for p in pkgs:
        dr.load_components(p, continue_on_error=False)


def apply_blacklist(cfg):

    def _skip_component(component):
        dr.set_enabled(component, enabled=False)
        blacklist.BLACKLISTED_SPECS.append(component.split('.')[-1])
        log.warning('WARNING: Skipping component: %s' % component)

    def _check_and_skip_component(name):
        def _isidentifier(name):
            if hasattr(str, 'isidentifier'):
                return name.isidentifier()
            else:
                return not any(it in name for it in ('/', ' ', '.', '-'))

        if not _isidentifier(name):
            # not symbolic name
            return False
        # like a symbolic name
        component = 'insights.specs.default.DefaultSpecs.{0}'.format(name)
        if dr.get_component_by_name(component):
            # is a sybmolic name
            _skip_component(component)
            return True
        # not symbolic name
        return False

    for fil in cfg.get("files", []):
        if not _check_and_skip_component(fil):
            blacklist.add_file(fil)

    for cmd in cfg.get("commands", []):
        if not _check_and_skip_component(cmd):
            blacklist.add_command(cmd)

    for component in cfg.get('components', []):
        if not dr.get_component_by_name(component):
            log.warning('WARNING: Unknown component in blacklist: %s' % component)
        else:
            _skip_component(component)

    if cfg.get('patterns'):
        log.warning("WARNING: Excluding patterns defined in blacklist configuration")

    if cfg.get('keywords'):
        log.warning("WARNING: Replacing keywords defined in blacklist configuration")


def create_context(ctx):
    """
    Loads and constructs the specified context with the specified arguments.
    If a '.' isn't in the class name, the 'insights.core.context' package is
    assumed.
    """
    ctx_cls_name = ctx.get("class", "insights.core.context.HostContext")
    if "." not in ctx_cls_name:
        ctx_cls_name = "insights.core.context." + ctx_cls_name
    ctx_cls = dr.get_component(ctx_cls_name)
    ctx_args = ctx.get("args", {})
    return ctx_cls(**ctx_args)


def get_to_persist(persisters):
    """
    Given a specification of what to persist, generates the corresponding set
    of components.
    """

    def specs():
        for p in persisters:
            if isinstance(p, dict):
                yield p["name"], p.get("enabled", True)
            else:
                yield p, True

    components = sorted(dr.DELEGATES, key=dr.get_name)
    names = dict((c, dr.get_name(c)) for c in components)

    results = set()
    for p, e in specs():
        for c in components:
            if names[c].startswith(p):
                if e:
                    results.add(c)
                elif c in results:
                    results.remove(c)
    return results


def create_archive(path, remove_path=True):
    """
    Creates a tar.gz of the path using the path basename + "tar.gz"
    The resulting file is in the parent directory of the original path, and
    the original path is removed.
    """
    root_path = os.path.dirname(path)
    relative_path = os.path.basename(path)
    archive_path = path + ".tar.gz"

    cmd = [["tar", "-C", root_path, "-czf", archive_path, relative_path]]
    call(cmd, env=SAFE_ENV)
    if remove_path:
        fs.remove(path)
    return archive_path


def generate_archive_name():
    """
    Creates the archive directory to store the component output.
    """
    hostname = determine_hostname()
    suffix = datetime.now(utc).strftime("%Y%m%d%H%M%S")
    return "insights-%s-%s" % (hostname, suffix)


def collect(
    client_config=None,
    rm_conf=None,
    tmp_path=None,
    archive_name=None,
    compress=False,
    manifest=None,
):
    """
    This is the collection entry point. It accepts a manifest, a temporary
    directory in which to store output, and a boolean for optional compression.

    Args:
        client_config (InsightsConfig): Configurations read from insights-client
            configuration, including "manifest".
        rm_conf (dict): Client-provided python dict containing keys
            "commands", "files", and "keywords", to be injected
            into the manifest blacklist.
        tmp_path (str): The temporary directory that is used to create a
            working directory for storing the final tar.gz if one is generated.
        archive_name (str): The directory that is used to generate the output
            as well as the final tar.gz.
        compress (boolean): True to create a tar.gz and remove the original
            workspace containing output. False to leave the workspace without
            creating a tar.gz
        manifest (str or dict): json document or dictionary containing the
            collection manifest. See default_manifest for an example.  This
            option works only for `insights-collect` where 'client_config'
            is not filled.

    Returns:
        (str, dict): The full path to the created tar.gz or workspace.
        And a dictionary with relevant exceptions captured by the broker during
        core collection, this dictionary has the following structure:
        ``{ exception_type: [ (exception_obj, component), (exception_obj, component) ]}``.
    """
    # Get the manifest per the following order:
    # 1. "client_config.manifest"
    # 2. "manifest" passed to the `insights.collect()`
    # 3. "default_manifest" defined in this module
    manifest = default_manifest if manifest is None else manifest
    if client_config and hasattr(client_config, 'manifest') and client_config.manifest:
        manifest = client_config.manifest
    manifest = load_manifest(manifest)

    client = manifest.get("client", {})
    plugins = manifest.get("plugins", {})

    load_packages(plugins.get("packages", []))
    apply_default_enabled(plugins)
    apply_configs(plugins)
    # process blacklist
    black_list = client.get("blacklist", {})
    black_list.update(rm_conf or {})
    apply_blacklist(black_list)

    # insights-client
    if client_config and client_config.cmd_timeout:
        try:
            client['context']['args']['timeout'] = client_config.cmd_timeout
        except LookupError:
            log.warning('Could not set timeout option.')

    try:
        filters.load()
    except IOError as e:
        # could not load filters file
        log.debug("No filters available: %s", str(e))
    except AttributeError as e:
        # problem parsing the filters
        log.debug("Could not parse filters: %s", str(e))

    output_path = os.path.join(tmp_path, archive_name)
    fs.ensure_path(output_path)
    fs.touch(os.path.join(output_path, "insights_archive.txt"))

    broker = dr.Broker()
    ctx = create_context(client.get("context", {}))
    cleaner = Cleaner(client_config, black_list) if client_config else None
    broker[ctx.__class__] = ctx
    broker['cleaner'] = cleaner
    broker['redact_config'] = black_list
    broker['client_config'] = client_config

    # run in "serial" mode by default
    run_strategy = client.get("run_strategy", {"name": "serial"})
    parallel = run_strategy.get("name") == "parallel"
    to_persist = get_to_persist(client.get("persist", set()))

    if client_config and client_config.obfuscate and parallel:
        log.warning("Parallel collection is not supported when 'obfuscate' is enabled")
        parallel = False

    pool_args = run_strategy.get("args", {})
    with get_pool(parallel, "insights-collector-pool", pool_args) as pool:
        h = Hydration(output_path, ctx, pool=pool)
        broker.add_observer(h.make_persister(to_persist))
        dr.run_all(broker=broker, pool=pool)

    collect_errors = _parse_broker_exceptions(broker, EXCEPTIONS_TO_REPORT)

    cleaner.generate_report(archive_name) if cleaner else None

    if compress:
        return create_archive(output_path), collect_errors
    return output_path, collect_errors


def _parse_broker_exceptions(broker, exceptions_to_report):
    """
    Parse the exceptions captured in the broker during core collection
    and keep only exception types configured in the ``exceptions_to_report``.

    Args:
        broker (Broker): Broker object used for core collection.
        exceptions_to_report (set): Exception types to retrieve from the broker.

    Returns:
        (dict): A dictionary with relevant exceptions captured by the broker.
    """
    errors = {}
    try:
        if broker.exceptions:
            for component, exceptions in broker.exceptions.items():
                for ex in exceptions:
                    ex_type = type(ex)
                    if ex_type in exceptions_to_report:
                        if ex_type in errors.keys():
                            errors[ex_type].append((ex, component))
                        else:
                            errors[ex_type] = [(ex, component)]
    except Exception as e:
        log.warning("Could not parse exceptions from the broker.: %s", str(e))
    return errors


def main():
    # Remove command line args so that they are not parsed by any called modules
    # The main fxn is only invoked as a cli, if calling from another cli then
    # use the collect function instead
    collect_args = [arg for arg in sys.argv[1:]] if len(sys.argv) > 1 else []
    sys.argv = (
        [
            sys.argv[0],
        ]
        if sys.argv
        else sys.argv
    )

    p = argparse.ArgumentParser()
    p.add_argument("-m", "--manifest", help="Manifest yaml.")
    p.add_argument("-o", "--out_path", help="Path to write output data.")
    p.add_argument("-q", "--quiet", help="Error output only.", action="store_true")
    p.add_argument("-v", "--verbose", help="Verbose output.", action="store_true")
    p.add_argument("-d", "--debug", help="Debug output.", action="store_true")
    p.add_argument("-c", "--compress", help="Compress", action="store_true")
    args = p.parse_args(args=collect_args)

    level = logging.WARNING
    if args.verbose:
        level = logging.INFO
    if args.debug:
        level = logging.DEBUG
    if args.quiet:
        level = logging.ERROR

    logging.basicConfig(level=level)

    out_path = args.out_path or tempfile.gettempdir()
    archive, errors = collect(
        manifest=args.manifest,
        tmp_path=out_path,
        archive_name=generate_archive_name(),
        compress=args.compress,
    )
    print(archive)


if __name__ == "__main__":
    main()



import pytest

from mock.mock import mock_open
from insights.combiners.ps import Ps
from insights.core.context import HostContext
from insights.core.exceptions import SkipComponent
from insights.parsers.mount import ProcMounts
from insights.specs.datasources.httpd import httpd_cmds, httpd_on_nfs, httpd_configuration_files, httpd24_scl_configuration_files, httpd24_scl_jbcs_configuration_files
from insights.tests import context_wrap


try:
    from unittest.mock import patch
    builtin_open = "builtins.open"
except Exception:
    from mock import patch
    builtin_open = "__builtin__.open"


MOUNT_DATA = """
/dev/mapper/root / ext4 rw,relatime,barrier=1,data=ordered 0 0
/dev/mapper/httpd1 /httpd1 nfs4 rw,relatime,vers=4,barrier=1,data=ordered 0 0
/dev/mapper/httpd2 /httpd2 nfs4 rw,relatime,vers=4,barrier=1,data=ordered 0 0
""".strip()

MOUNT_DATA_NO_NFS = """
/dev/mapper/root / ext4 rw,relatime,barrier=1,data=ordered 0 0
""".strip()

NFS_LSOF_666 = """
zsh     3520 httpd    3r   REG  253,0   6940392    648646 /httpd1
zsh     3520 httpd   11r   REG  253,0   9253600    648644 /httpd1
zsh     3520 httpd   12u   CHR    5,0       0t0      1034 /dev/tty
""".strip()
NFS_LSOF_777 = """
zsh     3520 httpd    3r   REG  253,0   6940392    648646 /httpd2
zsh     3520 httpd   11r   REG  253,0   9253600    648644 /httpd2
zsh     3520 httpd   12u   CHR    5,0       0t0      1034 /httpd2
""".strip()


class FakeContext(HostContext):
    def shell_out(self, cmd, split=True, timeout=None, keep_rc=False, env=None, signum=None):
        if 'pgrep' in cmd:
            return '666', '777'
        elif 'lsof -p 666' in cmd:
            return NFS_LSOF_666.splitlines()
        elif 'lsof -p 777' in cmd:
            return NFS_LSOF_777.splitlines()

        raise Exception


class FakeContext_NO_httpd(HostContext):
    def shell_out(self, cmd, split=True, timeout=None, keep_rc=False, env=None, signum=None):
        if 'pgrep' in cmd:
            return ''

        raise Exception


# The ``get_running_commands()`` is tested in:
# - insights/tests/datasources/test_get_running_commands.py
# Here, we do not test it
@patch('insights.specs.datasources.httpd.get_running_commands')
def test_httpd_cmds(run_cmds):
    broker = {Ps: None, HostContext: None}
    httpds = ['/usr/sbin/httpd/', '/opt/rh/httpd24/root/usr/sbin/httpd']
    run_cmds.return_value = httpds
    result = httpd_cmds(broker)
    assert result == httpds

    run_cmds.return_value = []
    with pytest.raises(SkipComponent):
        httpd_cmds(broker)


@patch('insights.specs.datasources.httpd.get_running_commands')
def test_httpd_on_nfs(run_cmds):
    broker = {ProcMounts: ProcMounts(context_wrap(MOUNT_DATA)), HostContext: FakeContext()}
    result = httpd_on_nfs(broker)
    assert '"http_ids": ["666", "777"]' in result.content[0]
    assert '"open_nfs_files": 5' in result.content[0]
    assert '"nfs_mounts": ["/httpd1", "/httpd2"]' in result.content[0]


@patch('insights.specs.datasources.httpd.get_running_commands')
def test_httpd_on_nfs_no_httpd(run_cmds):
    broker = {ProcMounts: ProcMounts(context_wrap(MOUNT_DATA)), HostContext: FakeContext_NO_httpd()}
    with pytest.raises(SkipComponent):
        httpd_on_nfs(broker)


@patch('insights.specs.datasources.httpd.get_running_commands')
def test_httpd_on_nfs_no_mount(run_cmds):
    broker = {ProcMounts: ProcMounts(context_wrap(MOUNT_DATA_NO_NFS)), HostContext: FakeContext()}
    with pytest.raises(SkipComponent):
        httpd_on_nfs(broker)


data_lines_httpd_conf = """
ServerRoot "/etc/httpd"
Include conf.d/*.conf
""".strip()

data_lines_httpd_conf_lower_case = """
ServerRoot "/etc/httpd"
include conf.d/*.conf
""".strip()

data_lines_httpd24_scl_conf = """
ServerRoot "/opt/rh/httpd24/root/etc/httpd"
Include conf.d/*.conf
""".strip()

data_lines_httpd24_scl_jbcs_conf = """
ServerRoot "/opt/rh/jbcs-httpd24/root/etc/httpd"
Include conf.d/*.conf
""".strip()

data_lines_ssl_conf = """
Listen 443 https
IncludeOptional modsecurity.d/*.conf
<IfModule mod_security2.c>
    # ModSecurity Core Rules Set and Local configuration
    IncludeOptional modsecurity.d/*.conf
</IfModule>
""".strip()

data_lines_ssl_conf_lower_case = """
Listen 443 https
includeOptional modsecurity.d/*.conf
<IfModule mod_security2.c>
    # ModSecurity Core Rules Set and Local configuration
    IncludeOptional modsecurity.d/*.conf
</IfModule>
""".strip()

data_lines_httpd_conf_section_test = """
ServerRoot "/etc/httpd"
<IfModule mod_security2.c>
    # ModSecurity Core Rules Set and Local configuration
    IncludeOptional modsecurity.d/*.conf
</IfModule>
""".strip()

data_lines_crs_setup_conf = """
SecAction \
    "id:900990,\
""".strip()


@patch("os.path.isfile", return_value=True)
@patch("os.path.isdir", return_value=True)
@patch("glob.glob", return_value=["/etc/httpd/conf.d/ssl.conf"])
@patch(builtin_open, new_callable=mock_open, read_data=data_lines_httpd_conf)
def test_httpd_conf_files(m_open, m_glob, m_isdir, m_isfile):
    handlers = (m_open.return_value, mock_open(read_data=data_lines_ssl_conf).return_value)
    m_open.side_effect = handlers
    broker = {HostContext: None}
    result = httpd_configuration_files(broker)
    assert result == set(['/etc/httpd/conf.d/ssl.conf', '/etc/httpd/conf/httpd.conf'])


@patch("os.path.isfile", return_value=True)
@patch("os.path.isdir", return_value=True)
@patch("glob.glob", return_value=["/etc/httpd/conf.d/ssl.conf"])
@patch(builtin_open, new_callable=mock_open, read_data=data_lines_httpd_conf_lower_case)
def test_httpd_conf_files_lower_case(m_open, m_glob, m_isdir, m_isfile):
    handlers = (m_open.return_value, mock_open(read_data=data_lines_ssl_conf_lower_case).return_value)
    m_open.side_effect = handlers
    broker = {HostContext: None}
    result = httpd_configuration_files(broker)
    assert result == set(['/etc/httpd/conf.d/ssl.conf', '/etc/httpd/conf/httpd.conf'])


@patch("os.path.isfile", return_value=True)
@patch("os.path.isdir", return_value=True)
@patch("glob.glob", return_value=["/etc/httpd/modsecurity.d/crs-setup.conf"])
@patch(builtin_open, new_callable=mock_open, read_data=data_lines_httpd_conf_section_test)
def test_httpd_conf_files_section(m_open, m_glob, m_isdir, m_isfile):
    handlers = (m_open.return_value, mock_open(read_data=data_lines_crs_setup_conf).return_value)
    m_open.side_effect = handlers
    broker = {HostContext: None}
    result = httpd_configuration_files(broker)
    assert result == set(['/etc/httpd/conf/httpd.conf'])


@patch("os.path.isfile", return_value=False)
@patch("os.path.isdir", return_value=False)
@patch("glob.glob", return_value=["/etc/httpd/conf.d/ssl.conf"])
@patch(builtin_open, new_callable=mock_open, read_data=data_lines_httpd_conf)
def test_httpd_conf_files_ssl_miss(m_open, m_glob, m_isdir, m_isfile):
    handlers = (m_open.return_value, mock_open(read_data=data_lines_ssl_conf).return_value)
    m_open.side_effect = handlers
    broker = {HostContext: None}
    result = httpd_configuration_files(broker)
    assert result == set(['/etc/httpd/conf/httpd.conf'])


@patch("os.path.join", return_value='/etc/httpd/no_such_file')
@patch("os.path.isfile", return_value=True)
@patch("os.path.isdir", return_value=True)
@patch("glob.glob", return_value=["/etc/httpd/conf.d/ssl.conf"])
def test_httpd_conf_files_main_miss(m_glob, m_isdir, m_isfile, m_join):
    broker = {HostContext: None}
    with pytest.raises(SkipComponent):
        httpd_configuration_files(broker)


@patch("os.path.isfile", return_value=True)
@patch("os.path.isdir", return_value=True)
@patch("glob.glob", return_value=["/opt/rh/httpd24/root/etc/httpd/conf.d/ssl.conf"])
@patch(builtin_open, new_callable=mock_open, read_data=data_lines_httpd24_scl_conf)
def test_httpd24_scl_conf_files(m_open, m_glob, m_isdir, m_isfile):
    handlers = (m_open.return_value, mock_open(read_data=data_lines_ssl_conf).return_value)
    m_open.side_effect = handlers
    broker = {HostContext: None}
    result = httpd24_scl_configuration_files(broker)
    assert result == set(['/opt/rh/httpd24/root/etc/httpd/conf.d/ssl.conf', '/opt/rh/httpd24/root/etc/httpd/conf/httpd.conf'])


@patch("os.path.isfile", return_value=True)
@patch("os.path.isdir", return_value=True)
@patch("glob.glob", return_value=["/opt/rh/jbcs-httpd24/root/etc/httpd/conf.d/ssl.conf"])
@patch(builtin_open, new_callable=mock_open, read_data=data_lines_httpd24_scl_jbcs_conf)
def test_httpd24_scl_jbs_conf_files(m_open, m_glob, m_isdir, m_isfile):
    handlers = (m_open.return_value, mock_open(read_data=data_lines_ssl_conf).return_value)
    m_open.side_effect = handlers
    broker = {HostContext: None}
    result = httpd24_scl_jbcs_configuration_files(broker)
    assert result == set(['/opt/rh/jbcs-httpd24/root/etc/httpd/conf.d/ssl.conf', '/opt/rh/jbcs-httpd24/root/etc/httpd/conf/httpd.conf'])



import doctest
import pytest

from insights.core.exceptions import SkipComponent, ParseException
from insights.parsers import ilab
from insights.parsers.ilab import IlabModuleList
from insights.tests import context_wrap

ILAB_MODULE_LIST = """
+-----------------------------------+---------------------+---------+
| Model Name                        | Last Modified       | Size    |
+-----------------------------------+---------------------+---------+
| models/prometheus-8x7b-v2-0       | 2024-08-09 13:28:50 |  87.0 GB|
| models/mixtral-8x7b-instruct-v0-1 | 2024-08-09 13:28:24 |  87.0 GB|
| models/granite-7b-redhat-lab      | 2024-08-09 14:28:40 |  12.6 GB|
| models/granite-7b-starter         | 2024-08-09 14:40:35 |  12.6 GB|
+-----------------------------------+---------------------+---------+
""".strip()

ILAB_MODULE_LIST_ISSUE_FORMAT = """
+-----------------------------------+---------------------+
| Model Name                        | Last Modified       |
+-----------------------------------+---------------------+
| models/mixtral-8x7b-instruct-v0-1 | 2024-08-09 13:28:24 |
| models/granite-7b-redhat-lab      | 2024-08-09 14:28:40 |
| models/granite-7b-starter         | 2024-08-09 14:40:35 |
+-----------------------------------+---------------------+
""".strip()

ILAB_MODULE_LIST_ERROR = """
Usage: ilab model [OPTIONS] COMMAND [ARGS]...
Try 'ilab model --help' for help.

Error: `/root/.config/instructlab/config.yaml` does not exist or is not a readable file.
Please run `ilab config init` or point to a valid configuration file using `--config=<path>`.
""".strip()

ILAB_MODULE_LIST_EMPTY = ""


def test_ilab_model_list():
    ilab_module_list_info = IlabModuleList(context_wrap(ILAB_MODULE_LIST))
    assert len(ilab_module_list_info) == 4
    assert ilab_module_list_info[0]["model_name"] == "models/prometheus-8x7b-v2-0"
    assert ilab_module_list_info[1]["last_modified"] == "2024-08-09 13:28:24"
    assert ilab_module_list_info[2]["size"] == "12.6 GB"
    assert ilab_module_list_info.models == ["models/prometheus-8x7b-v2-0", "models/mixtral-8x7b-instruct-v0-1", "models/granite-7b-redhat-lab", "models/granite-7b-starter"]

    with pytest.raises(SkipComponent):
        IlabModuleList(context_wrap(ILAB_MODULE_LIST_EMPTY))

    with pytest.raises(ParseException):
        IlabModuleList(context_wrap(ILAB_MODULE_LIST_ERROR))

    with pytest.raises(ParseException):
        IlabModuleList(context_wrap(ILAB_MODULE_LIST_ISSUE_FORMAT))


def test_ilab_doc_examples():
    env = {
        'ilab_model_list_info': IlabModuleList(context_wrap(ILAB_MODULE_LIST))
    }
    failed, total = doctest.testmod(ilab, globs=env)
    assert failed == 0



"""
Clean Specs (files/commands)
============================

The following modules are provided in the Cleaner and can be applied to the
specs during collection according to the user configuration and specs setting.

- Redaction (patterns redaction)
  This is a must-be-done operation to all the collected specs.  A `no_redact`
  option is available to specs, if it's surely contains non-security
  information, e.g. the `machine-id` spec.

- Filtering
  Filter lines as per the allow list got from the `filters.yaml`.  The
  `filtering` can only be applied when `allowlist` is available (not None) for
  the spec.

- Obfuscation (IPv4, [IPv6], Hostname, MAC, Password, Keywords)
  Obfuscate lines in spec content according to the user configuration and
  specs requirement.  The `no_obfuscate` can be used to exclude obfuscation
  target from the obfuscation.  Currently, the supported obfuscation target
  are:
  * hostname
  * ip (ipv4)
  * ipv6
  * keyword
  * mac
  * password
"""

import logging
import json
import os
import six

from insights.cleaner.filters import AllowFilter
from insights.cleaner.hostname import Hostname
from insights.cleaner.ip import IPv4  # IPv6
from insights.cleaner.keyword import Keyword

# from insights.cleaner.mac import Mac
from insights.cleaner.password import Password
from insights.cleaner.pattern import Pattern
from insights.cleaner.utilities import write_report
from insights.util.hostname import determine_hostname
from insights.util.posix_regex import replace_posix

logger = logging.getLogger(__name__)
MAX_LINE_LENGTH = 1048576  # 1MB
DEFAULT_OBFUSCATIONS = {
    'hostname',
    'ip',  # ipv4
    'ipv6',
    'keyword',
    'mac',
    'password',
}


class Cleaner(object):
    """
    Class to clean the content of Specs according to the user configuration and
    spec setting.
    """

    def __init__(self, config, rm_conf, fqdn=None):
        self.report_dir = '/tmp'  # FIXME
        self.rhsm_facts_file = getattr(
            config, 'rhsm_facts_file', os.path.join(self.report_dir, 'insights-client.facts')
        )
        # Handle User Configuration
        rm_conf = rm_conf or {}
        exclude = rm_conf.get('patterns', [])
        regex = False
        if isinstance(exclude, dict) and exclude.get('regex'):
            exclude = [r'%s' % replace_posix(i) for i in exclude['regex']]
            regex = True
        # - Pattern Redaction and allow-list Filtering
        self.redact = {
            'pattern': Pattern(exclude, regex) if exclude else None,
            'allow_filter': AllowFilter(),
        }
        # - Keyword and Password Replacement
        #   They Do NOT depend on "obfuscate=True"
        keywords = rm_conf.get('keywords')
        self.obfuscate = {
            'keyword': Keyword(keywords) if keywords else None,
            'password': Password(),
        }

        self.fqdn = fqdn if fqdn else determine_hostname()
        if config and config.obfuscate:
            # - IPv4 obfuscation
            self.obfuscate.update(ip=IPv4())
            # # - IPv6 obfuscation
            # self.obfuscate.update(ipv6=IPv6()) if config.obfuscate_ipv6 else None
            # - Hostname obfuscation
            (
                self.obfuscate.update(hostname=Hostname(self.fqdn))
                if config.obfuscate_hostname
                else None
            )
            # # - MAC obfuscation
            # self.obfuscate.update(mac=Mac()) if config.obfuscate_mac else None

    def clean_content(self, lines, no_obfuscate=None, no_redact=False, allowlist=None, width=False):
        """
        Clean lines one by one according to the configuration, the cleaned
        lines will be returned.
        """

        def _clean_line(line):
            if len(line) > MAX_LINE_LENGTH:
                # Keep the first MAX_LINE_LENGTH chars only (it rarely happens)
                line = line[:MAX_LINE_LENGTH]
                logger.debug('Extra-long line is truncated ...')

            for parser, kwargs in parsers:
                line = parser.parse_line(line, **kwargs)
            return line

        # List of parsers to be applied with Order
        parsers = list()
        # 1. Redact when NO "no_redact=True" is set
        if self.redact['pattern'] and not no_redact:
            parsers.append((self.redact['pattern'], {})) if not no_redact else None
        # 2. Filter as per allowlist got from add_filter
        (
            parsers.append((self.redact['allow_filter'], {'allowlist': allowlist}))
            if allowlist is not None
            else None
        )
        # 3. Obfuscation entries
        # - Hostname
        # - IPv4
        # - IPv6
        # - Keyword
        # - Mac
        # - Password
        no_obfuscate.append('ipv6') if no_obfuscate and 'ip' in no_obfuscate else None
        for obf in set(self.obfuscate.keys()) - set(no_obfuscate or []):
            if self.obfuscate[obf]:
                parsers.append((self.obfuscate[obf], {'width': width}))

        # handle single string
        if not isinstance(lines, list):
            return _clean_line(lines)

        result = []
        for line in lines:
            line = _clean_line(line)
            result.append(line) if line is not None else None
        if result and any(l for l in result):
            # When there are some lines Truth
            return result
        # All lines blank
        return []

    def clean_file(self, _file, no_obfuscate=None, no_redact=False, allowlist=None):
        """
        Clean a file according to the configuration, the file will be updated
        directly with the cleaned content.
        """
        logger.debug('Cleaning %s ...' % _file)

        if os.path.exists(_file) and not os.path.islink(_file):
            # Process the file
            raw_data = content = None
            try:
                with open(_file, 'r') as fh:
                    raw_data = fh.readlines()
                    content = self.clean_content(
                        raw_data,
                        no_obfuscate=no_obfuscate,
                        no_redact=no_redact,
                        allowlist=allowlist,
                        width=_file.endswith("netstat_-neopa"),
                    )
            except Exception as e:  # pragma: no cover
                logger.warning(e)
                raise Exception("Error: Cannot Open File for Cleaning: %s" % _file)
            # Store it
            try:
                if raw_data:
                    if content:
                        with open(_file, 'wb') as fh:
                            for line in content:
                                fh.write(line.encode('utf-8') if six.PY3 else line)
                    else:
                        # Remove Empty file
                        logger.debug('Removing %s, as it\'s empty after cleaning' % _file)
                        os.remove(_file)
            except Exception as e:  # pragma: no cover
                logger.warning(e)
                raise Exception("Error: Cannot Write to File: %s" % _file)

    def generate_rhsm_facts(self):
        logger.info('Writing RHSM facts to %s ...', self.rhsm_facts_file)

        hostname = self.obfuscate.get('hostname')
        hn_mapping = hostname.mapping() if hostname else []

        keyword = self.obfuscate.get('keyword')
        kw_mapping = keyword.mapping() if keyword else []

        ipv4 = self.obfuscate.get('ip')
        ipv4_mapping = ipv4.mapping() if ipv4 else []

        facts = {
            'insights_client.hostname': self.fqdn,
            'insights_client.obfuscate_ip_enabled': 'ip' in self.obfuscate,
            # 'insights_client.obfuscate_ipv6_enabled': 'ipv6' in self.obfuscate,
            # 'insights_client.obfuscate_mac_enabled': 'mac' in self.obfuscate,
            'insights_client.obfuscate_hostname_enabled': 'hostname' in self.obfuscate,
            'insights_client.obfuscated_ipv4': json.dumps(ipv4_mapping),
            # 'insights_client.obfuscated_ipv6': json.dumps(),
            # 'insights_client.obfuscated_mac': json.dumps(),
            'insights_client.obfuscated_keyword': json.dumps(kw_mapping),
            'insights_client.obfuscated_hostname': json.dumps(hn_mapping),
        }

        write_report(facts, self.rhsm_facts_file)

    def generate_report(self, archive_name):
        # Always generate the rhsm.facts files
        self.generate_rhsm_facts()
        # Generate CSV reports accordingly
        for parser in list(self.redact.values()) + list(self.obfuscate.values()):
            if parser:
                parser.generate_report(self.report_dir, archive_name)



"""
Custom datasources to get dict of file or dir numbers
"""
import json
import os
from insights.core.context import HostContext
from insights.core.exceptions import SkipComponent
from insights.core.filters import get_filters
from insights.core.plugins import datasource
from insights.core.spec_factory import DatasourceProvider
from insights.specs import Specs


@datasource(HostContext)
def files_dirs_number(broker):
    """ Return a dict of file numbers from the spec filter """
    filters = sorted(get_filters(Specs.files_dirs_number_filter))
    result = {}
    if filters:
        for item in filters:
            if not item.endswith("/"):
                item = item + "/"
            if os.path.exists(item):
                files_number = 0
                dirs_number = 0
                result[item] = {}
                for name in os.listdir(item):
                    if os.path.isfile(item + name):
                        files_number = files_number + 1
                    elif os.path.isdir(item + name):  # pragma: no cover
                        dirs_number = dirs_number + 1
                result[item]["files_number"] = files_number
                result[item]["dirs_number"] = dirs_number
        if result:
            return DatasourceProvider(content=json.dumps(result, sort_keys=True), relative_path='files_dirs_number')
    raise SkipComponent



"""
Rpm -V Packages Parsers - command ``/bin/rpm -V <packages>``
============================================================

Parsers provided in this module includes:

RpmVPackage - command ``/bin/rpm -V <package>``
-----------------------------------------------
"""

from insights.core import CommandParser
from insights.core.exceptions import SkipComponent
from insights.core.plugins import parser
from insights.specs import Specs


@parser(Specs.rpm_V_package)
class RpmVPackage(CommandParser):
    """
    Class for parsing ``/bin/rpm -V <package>`` command.

    Attributes:
        package_name (String): The name of package
        discrepancies (list of dictionaries): Every dictionary contains information about one entry

    Sample output of this command is::

        ..?......  c /etc/sudoers
        ..?......    /usr/bin/sudo
        ..?......    /usr/bin/sudoreplay
        missing     /var/db/sudo/lectured (Permission denied)

    Examples:
        >>> type(rpm_v_pkg)
        <class 'insights.parsers.rpm_v_packages.RpmVPackage'>
        >>> rpm_v_pkg.package_name
        'sudo'
        >>> len(rpm_v_pkg.discrepancies)
        4
        >>> sorted(rpm_v_pkg.discrepancies[0].items())
        [('attributes', '..?......'), ('file', '/etc/sudoers'), ('line', '..?......  c /etc/sudoers'), ('mark', 'c')]
    """

    def parse_content(self, content):
        self.discrepancies = []
        self.package_name = self.file_name.split('_')[-1] if self.file_name else None

        for line in content:
            line_parts = line.split()
            if not line_parts:
                continue

            if "package" in line_parts[0]:
                raise SkipComponent("Invalid Contents")

            if "missing" in line_parts[0]:
                entry = {"line": line.strip(), "attributes": None, "mark": None, "file": None}
            elif len(line_parts) == 3:
                entry = {
                    "line": line.strip(),
                    "attributes": line_parts[0],
                    "mark": line_parts[1],
                    "file": line_parts[2],
                }
            else:
                entry = {
                    "line": line.strip(),
                    "attributes": line_parts[0],
                    "mark": None,
                    "file": line_parts[1],
                }
            self.discrepancies.append(entry)

        if not self.discrepancies:
            raise SkipComponent("Empty result")
